#!/usr/bin/env python
# AutoSwitcher - By QB89Dragon <hughescih@hotmail.com> 2011
# Automatically switches display to external display for laptops
# Tested on MacBookPro5,5 with VGA
# Uses the Disper libraries as a backend

import os
import sys
import gobject
import switcher
import pynotify
import optparse
import gtk

show_appindicator = True
try:
	import appindicator
except:
	show_appindicator = False
	
"""
	ToDo:
		- Notification Priorities
			n.set_urgency(pynotify.URGENCY_LOW)
			n.set_urgency(pynotify.URGENCY_NORMAL)
			n.set_urgency(pynotify.URGENCY_CRITICAL)
"""

def daemonize (stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
    # Perform first fork.
    try:
        pid = os.fork( )
        if pid > 0:
            sys.exit(0) # Exit first parent.
    except OSError, e:
        sys.stderr.write("fork #1 failed: (%d) %sn" % (e.errno, e.strerror))
        sys.exit(1)
    # Decouple from parent environment.
    os.chdir("/")
    os.umask(0)
    os.setsid( )
    # Perform second fork.
    try:
        pid = os.fork( )
        if pid > 0:
            # Write the pid to the default PID file
            #o = open(PID_FILE, 'w')
            #o.write(str(pid))
            #o.close()
            sys.exit(0) # Exit second parent.
    except OSError, e:
        sys.stderr.write("fork #2 failed: (%d) %sn" % (e.errno, e.strerror))
        sys.exit(1)
    # The process is now daemonized, redirect standard file descriptors.
    for f in sys.stdout, sys.stderr: f.flush( )
    si = file(stdin, 'r')
    so = file(stdout, 'a+')
    se = file(stderr, 'a+', 0)
    os.dup2(si.fileno( ), sys.stdin.fileno( ))
    os.dup2(so.fileno( ), sys.stdout.fileno( ))
    os.dup2(se.fileno( ), sys.stderr.fileno( ))


class ModeSwitchAutomator:
	def __init__(self, menu=None, debug=False):
		self.menu = menu
		self.poll_delay = 2000
		self.sw = switcher.Switcher()
		self.debug = debug
		
		# Define defaults
		self.mode = 'external'
		self.direction = 'left'
		
		# Load settings
		try:
			f = open(os.path.expanduser('~')+'/.autoswitcher', 'r')
			data = f.read()
			for line in data.split('\n'):
				exec(line)
			f.close()
		except:
			pass
		
		if self.menu:
			self.menu.set_checked(self.mode)
			self.menu.set_direction_checked(self.direction)
			
			self.menu.connect('change', self.user_changed)
			self.menu.connect('change_direction', self.user_changed_direction)
		
		# Switch modes on startup if second display present by only listing the
		#  primary display on startup so at the first poll it sees the second display as new
		self.cached_display_list = self.listDisplays()
		self.cached_resolution_list = self.getDisplayResolutions()
		
		if len(self.cached_display_list) > 1:
			self.menu.show()
			self.set_display_mode(self.mode, self.direction)
			added_displays = self.cached_display_list - set([self.sw.get_primary_display()])
			for display in added_displays:
				self.display_added(display)
			
		if self.debug:
			print "Initially Configured Displays", self.cached_display_list
			print "Resolution", self.cached_resolution_list

	def listDisplays(self):
		return set(self.sw.get_displays(reprobe=True))
	
	def getDisplayNames(self):
		screens = {}
		for disp in self.sw.get_displays(reprobe=True):
			screens[disp] = {'name':self.sw.get_display_name(disp)}
		return screens

	def getDisplays(self):
		screens = {}
		for disp in self.sw.get_displays(reprobe=True):
			res = self.sw.get_resolutions_display(disp, reprobe=True)
			res.sort()
			screens[disp] = {'name':self.sw.get_display_name(disp), 'nativeResolution':res[-1], 'resolutionList':res}
		return screens
	
	def getDisplayResolutions(self):
		displays = self.getDisplays()
		resolutions = {}
		for display in displays:
			resolutions[display] = displays[display]['nativeResolution']
		return resolutions
		
	def startDisplayPoller(self):
		gobject.timeout_add(self.poll_delay, self.pollForDisplays)

	def pollForDisplays(self):
		try:
			displays = self.listDisplays()
			# Check for added displays
			added_displays = displays - self.cached_display_list
			for display in added_displays:
				if self.debug: print "Display Added: %s"%display
				self.display_added(display)
			# Check for removed displays
			removed_displays = self.cached_display_list - displays
			for display in removed_displays:
				if self.debug: print "Display Removed: %s"%display
				self.display_removed(display)
			self.cached_display_list = displays
			# Check for altered displays
			# N.B. This code is non-functional for resolution degredation. Highest resolution will be cached by the backend.
			resolutions = self.getDisplayResolutions()
			for display in resolutions:
				if display in self.cached_resolution_list:
					if str(resolutions[display]) != str(self.cached_resolution_list[display]):
						if self.debug: print "Display Changed: %s"%display
						self.display_added(display)
			self.cached_resolution_list = resolutions
		except:
			print "Unexpected Error %s"%str(sys.exc_info()[0])
		return True

	def set_display_mode(self, mode, data=None):
		if mode == 'external':
			
			displays = self.listDisplays() - set([self.sw.get_primary_display()])
			displays_data = self.getDisplays()
			del displays_data[self.sw.get_primary_display()]
			
			resolutions = [displays_data[display]['nativeResolution'] for display in displays]
			# Assumes all external monitors of same resolution as last in list.
			resolution = resolutions.pop()
			
			if self.debug: print "Display Resolution Detected:", resolution
			self.sw.switch_clone(displays, resolution)
			self.sw.set_scaling(displays, "default")
			
		elif mode == 'internal':
			
			display = self.sw.get_primary_display()
			resolution = self.getDisplays()[display]['nativeResolution']
			display_name = self.getDisplayNames()[display]['name']

			if self.debug: print "Display Resolution Detected:", resolution
			
			self.sw.switch_clone([display], resolution)
			self.sw.set_scaling([display], "default")

		elif mode == 'clone':
			
			displays = self.listDisplays()
			displays_data = self.getDisplays()
			resolutions = self.sw.get_resolutions(displays).common()
			if not resolutions:
				print "Displays share no common resolution, cannot clone"
				return
			resolution = sorted(resolutions)[-1]

			if self.debug: print "Display Resolution Detected:", resolution
			# The following line seemingly overrides a bug in cloning displays using the switcher module
			self.set_display_mode('external')

			self.sw.switch_clone(displays, resolution)
			self.sw.set_scaling(displays, "default")
			
		elif mode == 'extend':
			displays = self.listDisplays()
			displays_data = self.getDisplays()
			
			resolutions = {}
			for display in displays:
				resolutions[display] = displays_data[display]['nativeResolution']
				
			if self.debug: print "Display Resolutions Detected:", resolutions
			self.sw.switch_extend(displays, data, resolutions)
			self.sw.set_scaling(displays, "default")
			
	def notify(self, text, description):
		if pynotify.init("autoswitcher"):
			n = pynotify.Notification(text, description, "monitor")
			n.show()

	def user_changed(self, ev, mode):
		self.mode = mode
		self.set_display_mode(self.mode, self.direction)
		self.write_settings_file()
		
	def user_changed_direction(self, ev, direction):
		self.direction = direction
		if self.mode == 'extend':
			self.set_display_mode(self.mode, self.direction)
		self.write_settings_file()

	def write_settings_file(self):
		f = open(os.path.expanduser('~')+'/.autoswitcher', 'w')
		f.write('self.mode="'+self.mode+'"\n')
		f.write('self.direction="'+self.direction+'"\n')
		f.close()
		

	def display_added(self, display):
		if self.menu: self.menu.show()
		display_name = self.getDisplayNames()[display]['name']
		self.notify("External Display","Monitor %s will now be enabled"%display_name)
		self.set_display_mode(self.mode, self.direction)

	def display_removed(self, display):
		if self.menu: self.menu.hide()
		self.notify("External Display Removed","Primary display will now be used")
		self.set_display_mode('internal')

class AppIndicator(gobject.GObject):
	__gsignals__={	"change": (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, [gobject.TYPE_STRING]),
					"change_direction": (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, [gobject.TYPE_STRING])}
	
	def __init__(self):
		gobject.GObject.__init__(self)
		self.ind = appindicator.Indicator ("autoswitcher","gsd-xrandr", appindicator.CATEGORY_APPLICATION_STATUS)
		self.ind.set_status(appindicator.STATUS_PASSIVE)

		self.menu = gtk.Menu()
		self.menu_item_tags = ['external', 'internal', 'clone', 'extend']
		self.menu_items = {}
		primary_menu_item = gtk.RadioMenuItem(None, "External only")
		self.menu_items['external'] = primary_menu_item
		self.menu_items['internal'] = gtk.RadioMenuItem(primary_menu_item, "Laptop only")
		self.menu_items['clone'] = gtk.RadioMenuItem(primary_menu_item, "Clone desktop")
		self.menu_items['extend'] = gtk.RadioMenuItem(primary_menu_item, "Extend desktop")

		self.menu_items['external'].set_active(True)

		for item in self.menu_item_tags: 
			self.menu_items[item].show()
			self.menu.append(self.menu_items[item])
			self.menu_items[item].connect('activate', lambda w,t: self.emit('change', t), item)

		direction = gtk.MenuItem("Extend to")
		direction_menu = gtk.Menu()
		direction_menu_tags = ['left', 'right', 'top', 'bottom']
		l = {'left':'Left', 'right':'Right', 'top':'Top', 'bottom':'Bottom'}
		self.direction_menu_items = {}
		direction_primary = None
		for tag in direction_menu_tags:
			ri = gtk.RadioMenuItem(direction_primary, l[tag])
			self.direction_menu_items[tag] = ri
			if not direction_primary:
				direction_primary = ri
				ri.set_active(True)
			ri.show()
			direction_menu.append(ri)
			ri.connect('activate', lambda w,t: self.emit('change_direction', t), tag)


		direction.set_submenu(direction_menu)
		direction.show()
		self.menu.append(direction)
		self.ind.set_menu(self.menu)
	
	def show(self):
		self.ind.set_status(appindicator.STATUS_ACTIVE)

	def hide(self):
		self.ind.set_status(appindicator.STATUS_PASSIVE)

	def set_checked(self, tag):
		self.menu_items[tag].set_active(True)
	
	def set_direction_checked(self, tag):
		self.direction_menu_items[tag].set_active(True)


if __name__=="__main__":
	usage = "usage: %prog [options] arg"
	parser = optparse.OptionParser(usage)
	parser.add_option("-v", "--verbose", action="store_true", dest="verbose")
	parser.add_option("-d", "--no-daemon", action="store_false", dest="daemonize")
	(options, args) = parser.parse_args()
	if not options.verbose: options.verbose = False
	if options.daemonize == None: options.daemonize = True
	if options.daemonize: daemonize()
	if show_appindicator:
		menu = AppIndicator()
	autoModeSwitcher = ModeSwitchAutomator(menu, debug=options.verbose)
	autoModeSwitcher.startDisplayPoller()
	#gobject.MainLoop().run()
	gtk.main()
	
